<!DOCTYPE html>
<html>
<head>
<meta charset="utf-8" />
<meta name="viewport" content="width=device-width, initial-scale=1" />
{% include "metadata.html" %}{% if css %}<style>
{{ css|safe }}
</style>
{% endif %}</head>
<body>
<div id="root">{{ html|safe if html else "" }}</div>
<script type="module">
{{ js|safe }}
</script>
</body>
</html>
